import itertools
import os
import pprint
from typing import List, Dict

from immutabledict import immutabledict
from requests import Session

from fleabay.definitions import ItemSummary, ItemDetails, MultiItemModelAxis
from fleabay.utils import chunker, decimal_price_to_pence


class EbayApiExtractor:
    def __init__(self, api_key: str):
        self.api_key = api_key
        self.session = Session()

    def search(self, term: str, items_per_page: int = 200, remote_lowest_first: bool = False) -> List[ItemSummary]:
        url = "https://svcs.ebay.com/services/search/FindingService/v1"
        params = {
            "SECURITY-APPNAME": self.api_key,
            "OPERATION-NAME": "findItemsByKeywords",
            "SERVICE-VERSION": "1.0.0",
            "RESPONSE-DATA-FORMAT": "JSON",
            "REST-PAYLOAD": "",
            "keywords": term,
            "paginationInput.entriesPerPage": items_per_page,
            "GLOBAL-ID": "EBAY-GB",
            "siteid": 3,
            "buyerPostalCode": "M11 2GX"
        }
        if remote_lowest_first:
            params["sortOrder"] = "PricePlusShippingLowest"

        resp = self.session.get(url, params=params).json()
        resp_k = resp["findItemsByKeywordsResponse"][0]
        if resp_k["ack"][0] != "Success":
            raise RuntimeError("search by API not successful.")

        items = []

        found_item_ids = dict()

        if resp_k["searchResult"][0]["@count"] == "0":
            return []

        try:
            item_results = resp_k["searchResult"][0]["item"]
        except KeyError as e:
            raise RuntimeError("%r" % resp_k["searchResult"][0].items()) from e

        for result in item_results:
            item_id = result["itemId"][0]

            if item_id in found_item_ids:
                continue

            if item_id == "254817515397":
                pprint.pprint(result)

            is_buy_it_now = result["listingInfo"][0]["listingType"][0] != "Auction"

            postage_price = result["shippingInfo"][0]["shippingServiceCost"][0]
            postage_pence = -42
            if postage_price["@currencyId"] == "GBP":
                postage_pence = int(100 * float(postage_price["__value__"]))

            if result["shippingInfo"][0]["shippingType"][0] == "FreePickup":
                postage_pence = None

            item_price = result["sellingStatus"][0]["currentPrice"][0]
            item_pence = -42

            bid_count = result["sellingStatus"][0].get("bidCount")
            if bid_count is not None:
                bid_count = bid_count[0]

            if item_price["@currencyId"] == "GBP":
                item_pence = int(100 * float(item_price["__value__"]))

            view_url = f"https://www.ebay.co.uk/itm/{item_id}"

            try:
                condition = result["condition"][0]["conditionDisplayName"][0]
            except KeyError:
                condition = "???"

            if item_id == "154280943581":
                pprint.pprint(result)

            item = ItemSummary(
                item_id, view_url, result["title"][0], condition,
                item_pence, -42, None if is_buy_it_now else bid_count, result["country"][0], postage_pence, result["galleryURL"][0]
            )

            found_item_ids[item_id] = item

            items.append(item)

        for item_id, details in self.details_batch(list(found_item_ids.keys())).items():
            summary = found_item_ids[item_id]
            summary.extras["details"] = details
            summary.min_price = int(details.extra["min_price"])
            summary.max_price = int(details.extra["max_price"])

        return items

    @staticmethod
    def details(summary: ItemSummary) -> ItemDetails:
        return summary.extras["details"]

    def details_batch(self, ids: List[str]) -> Dict[str, ItemDetails]:
        # eBay can do up to 20 items at once.
        result = dict()
        for id_chunk in chunker(ids, 20):
            url = "https://open.api.ebay.com/shopping"
            params = {
                "ItemID": ",".join(id_chunk),
                "callname": "GetMultipleItems",
                "IncludeSelector": "Details,Variations",
                "appid": self.api_key,
                "siteid": 3,
                "version": "967",
                "responseencoding": "JSON"
            }
            response = self.session.get(url, params=params)
            # print(response.status_code)

            resp = response.json()

            for item_dict in resp["Item"]:
                item_id = item_dict["ItemID"]
                seller_dict = item_dict["Seller"]
                seller = {
                    "name": seller_dict["UserID"],
                    "feedback": f"{seller_dict['PositiveFeedbackPercent']}% ({seller_dict['FeedbackScore']})"
                }

                extras = {
                    "variant_extras": dict()
                }

                base_pence = -42
                if item_dict["CurrentPrice"]["CurrencyID"] == "GBP":
                    base_pence = decimal_price_to_pence("%.2f" % item_dict["CurrentPrice"]["Value"])

                min_pence = base_pence
                max_pence = base_pence

                variants_dict = dict()
                axes: Dict[str, MultiItemModelAxis] = dict()
                axes_inverse: Dict[str, Dict[str, int]] = dict()

                variations = item_dict.get("Variations")
                if variations is not None:
                    min_pence = 1e9
                    max_pence = -42

                    for axis in variations["VariationSpecificsSet"]["NameValueList"]:
                        axes_inverse[axis["Name"]] = {
                            val: i for i, val in enumerate(axis["Value"])
                        }
                        axes[axis["Name"]] = MultiItemModelAxis(axis["Name"], dict(enumerate(axis["Value"])))

                    for variation in variations["Variation"]:
                        quantity = variation["Quantity"] - variation["SellingStatus"].get("QuantitySold", 0)

                        if quantity <= 0:
                            continue

                        price = -42
                        if variation["StartPrice"]["CurrencyID"] == "GBP":
                            price = decimal_price_to_pence("%.2f" % variation["StartPrice"]["Value"])

                            min_pence = min(min_pence, price)
                            max_pence = max(max_pence, price)

                        sold = variation["SellingStatus"]["QuantitySold"]

                        variant_immdict = immutabledict({
                            pair["Name"]: axes_inverse[pair["Name"]][pair["Value"][0]]
                            for pair in variation["VariationSpecifics"]["NameValueList"]
                        })

                        extras["variant_extras"][variant_immdict] = {
                            "quantity": quantity,
                            "sold": sold
                        }

                        assert isinstance(price, int)
                        variants_dict[variant_immdict] = price

                        if item_id == "223749865106":
                            pprint.pprint(variation)
                            print("vi", "%r" % variant_immdict)
                            print("vp", "%r" % price)
                            print("qt", "%r" % quantity)

                extras["min_price"] = int(min_pence)
                extras["max_price"] = int(max_pence)

                result[item_id] = ItemDetails(
                    seller_dict["UserID"],
                    int(seller_dict["FeedbackScore"]),
                    seller_dict["PositiveFeedbackPercent"],
                    axes,
                    variants_dict,
                    extras
                )

        return result
