import ast
import json
from pprint import pprint
from re import Match
from typing import List, Dict

import bs4
from bs4 import Tag, NavigableString
from immutabledict import immutabledict
from requests import Session
import re

from fleabay.definitions import ItemSummary, ItemDetails, MultiItemModelAxis


class EbayExtractor:
    def __init__(self, domain: str = "https://www.ebay.co.uk"):
        self.domain = domain
        self.session = Session()

    def search(self, term: str, items_per_page: int = 200, remote_lowest_first: bool = False) -> List[ItemSummary]:
        params = {
            "_nkw": term,
            "_sacat": 0,
            "_ipg": items_per_page,
            # foreign country ID = 3 (UK)
            "_fcid": 3,
            # postcode
            "_stpos": "M11 2GX"
        }

        if remote_lowest_first:
            params["_sop"] = 15

        resp = self.session.get(self.domain + "/sch/i.html", params=params)

        if resp.status_code != 200:
            raise RuntimeError("not 200")

        doc = bs4.BeautifulSoup(resp.content, "html.parser")

        results = []

        item_tag: Tag
        for item_tag in doc.select(".s-item"):
            title_tag = item_tag.select_one(".s-item__title")
            if title_tag is None:
                continue
            for child in title_tag.children:
                if isinstance(child, NavigableString):
                    title = str(child)
                    break
            else:
                continue

            condition_tag = item_tag.select_one(".SECONDARY_INFO")
            if condition_tag:
                condition = condition_tag.text
            else:
                condition = "?"

            price_tag = item_tag.select_one(".s-item__price")
            if price_tag:
                prices = re.findall(r"£([0-9]+)\.([0-9]+)", price_tag.text)
                if not prices:
                    print("No £ in ", price_tag.text)
                    continue
                min_price_m = prices[0]
                max_price_m = prices[-1]
                min_price = int(min_price_m[0]) * 100 + int(min_price_m[1])
                max_price = int(max_price_m[0]) * 100 + int(max_price_m[1])
            else:
                min_price = -42
                max_price = -42

            bid_tag = item_tag.select_one(".s-item__bidCount")
            if bid_tag:
                bids = int(re.match("[0-9]+", bid_tag.text).group(0))
            else:
                bids = None

            postage_tag = item_tag.select_one(".s-item__logisticsCost")
            if postage_tag:
                if "Free" in postage_tag.text:
                    postage = 0
                else:
                    pmatch = re.search(r"£([0-9]+)\.([0-9]+)", postage_tag.text)
                    if pmatch is None:
                        print("no post ", postage_tag.text, item_tag)
                    postage = int(pmatch.group(1)) * 100 + int(pmatch.group(2))
            elif item_tag.select_one(".s-item__localDelivery") is not None:
                # local collection only.
                postage = None
            else:
                raise ValueError("Can't find postage in " + str(item_tag))

            location_tag = item_tag.select_one(".s-item__itemLocation")
            location = ""
            if location_tag and location_tag.text.startswith("From "):
                location = location_tag.text[5:]

            link_tag = item_tag.find("a")
            link_href = link_tag.attrs["href"]

            image_tag = item_tag.find("img")
            image_src = image_tag.attrs["src"]

            results.append(
                ItemSummary(
                    link_href,
                    link_href,
                    title,
                    condition,
                    min_price,
                    max_price,
                    bids,
                    location,
                    postage,
                    image_src,
                )
            )

        return results

    def details(self, summary: ItemSummary) -> ItemDetails:
        page = self.session.get(summary.href)
        line: bytes
        for line in page.iter_lines():
            opening = line.find(b"$rwidgets([")
            if opening == -1:
                continue
            ending = line.find(b"])", opening)
            if ending == -1:
                raise ValueError("Unable to find ending.")
            break
        else:
            raise ValueError("Unable to find opening.")

        pythonic_literal = (
            line[opening + 10 : ending + 1]
            .decode()
            .replace(":null", ":None")
            .replace(":false", ":False")
            .replace(":true", ":True")
        )
        data = ast.literal_eval(pythonic_literal)

        # pprint(data)
        # print("-----")

        for item in data:
            if item[0] == "com.ebay.raptor.vi.lockedheader.LockedHeaderCore":
                root_dict = item[2]
                variations = root_dict["itemVariationsModel"]["itemVariationsMap"]
                break
        else:
            raise ValueError("Variations not found")

        # pprint(variations)

        item_map: Dict[int, str] = {
            int(k): v["displayName"]
            for k, v in root_dict["itemVariationsModel"]["menuItemMap"].items()
        }
        models = root_dict["itemVariationsModel"]["menuModels"]
        prices: Dict[immutabledict, int] = dict()  # id[str, int]

        # pprint(models)

        for variation in variations.values():
            pmatch = re.search(r"£([0-9]+)\.([0-9]+)", variation["convertedPrice"])
            if pmatch is None:
                raise ValueError(
                    "Unable to extract price from " + variation["convertedPrice"]
                )
            price = int(pmatch.group(1)) * 100 + int(pmatch.group(2))

            trait_map = immutabledict(variation["traitValuesMap"])
            prices[trait_map] = price

        axes = dict()
        for model in models:
            value_map = dict()
            for value in model["menuItemValueIds"]:
                value_map[value] = item_map[value]

            axes[model["name"]] = MultiItemModelAxis(model["name"], value_map)

        return ItemDetails("", -1, "-1%", axes, prices)


if __name__ == "__main__":
    # EbayExtractor().search("fish tank", items_per_page=10)

    pprint(
        EbayExtractor().details(
            "https://www.ebay.co.uk/itm/Genuine-Perspex-Acrylic-Sheet-Black-White-Panel-Cut-To-Size-Plastic-3mm-5mm/124449453991?var=425326090110&hash=item1cf9c3f7a7:g:MPgAAOSw42Nft55q"
        )
    )
