import asyncio
import os
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Dict, Any, Tuple

from quart import Quart, render_template, request
from wtforms_async import Form, StringField, BooleanField, SelectField

from fleabay.extractors.ebay_api_extractor import EbayApiExtractor
from fleabay.definitions import ItemDetails
from fleabay.extractors.ebay_raw_extractor import EbayExtractor
from fleabay.extractors.zapiex_api_extractor import ZapiexApiExtractor

app = Quart(__name__)

pool = ThreadPoolExecutor(8)


def make_extractors() -> Tuple[Dict[str, Any], Dict[str, str]]:
    available_extractors = dict()
    extractor_names = dict()

    ebay_key = os.environ.get("EBAY_KEY")
    if ebay_key:
        available_extractors["eBay_api"] = EbayApiExtractor(ebay_key)
        extractor_names["eBay_api"] = "eBay (by API — best)"

    available_extractors["eBay_scrape"] = EbayExtractor()
    extractor_names["eBay_scrape"] = "eBay (by scraper — slow, error-prone)"

    zapiex_key = os.environ.get("ZAPIEX_KEY")
    if zapiex_key:
        available_extractors["AE_zapiex"] = ZapiexApiExtractor(zapiex_key)
        extractor_names["AE_zapiex"] = "AliExpress (by Zapiex API — limited)"

    return available_extractors, extractor_names


# we need only initialise the extractors once.
available_extractors, extractor_names = make_extractors()


class SearchForm(Form):
    extractor = SelectField("Site/Extractor", choices=list(extractor_names.items()))
    search_term = StringField("Search Term")
    include_local_collection = BooleanField("Include local collection")
    remote_lowest_first = BooleanField("Ask remote for 'Lowest Price+P&P First'")


@app.route("/", methods=["GET"])
async def index():
    return await render_template("index.html.j2", search_form=SearchForm())


@app.route("/search", methods=["GET"])
async def search():
    search_form = SearchForm(request.args)

    if not await search_form.validate():
        return await render_template("index.html.j2", search_form=SearchForm())

    ex = available_extractors[search_form.extractor.data]

    search_term = search_form.search_term.data

    loop = asyncio.get_running_loop()

    results = await loop.run_in_executor(pool, ex.search, search_term, 100, search_form.remote_lowest_first.data)

    if not search_form.include_local_collection.data:
        results = [r for r in results if r.postage is not None]

    item_details_fut = dict()

    for result in results:
        if not isinstance(ex, EbayExtractor) or result.min_price != result.max_price:
            fut = loop.run_in_executor(pool, ex.details, result)
            item_details_fut[fut] = result.id

    futs = tuple(item_details_fut.keys())
    detail_results = await asyncio.gather(*futs, return_exceptions=True)

    item_details: Dict[str, ItemDetails] = {
        item_details_fut[fut]: res
        for res, fut in zip(detail_results, futs)
        if isinstance(res, ItemDetails)
    }
    axes_errors = {
        item_details_fut[fut]: str(res)
        for res, fut in zip(detail_results, futs)
        if isinstance(res, Exception)
    }

    print(f"{len(axes_errors)} errors")

    # want to find the most insightful axis to display over
    # Insightful suggests that the individual values in the axis have a tight range
    axes_ranges = dict()
    for item_id, details in item_details.items():
        best_axis = None
        best_axis_tightness = 1.0e9
        for axis in details.axes.values():
            tightnesses = []
            for value, _v_name in axis.values.items():
                min_vp = 1e9
                max_vp = 0
                for variant, price in details.variants.items():
                    if variant[axis.code_name] != value:
                        continue
                    max_vp = max(max_vp, price)
                    min_vp = min(min_vp, price)

                tightnesses.append(max_vp - min_vp)
            avg_tightness = sum(tightnesses) / len(tightnesses)
            if avg_tightness < best_axis_tightness:
                best_axis = axis
                best_axis_tightness = avg_tightness

        if best_axis is not None:
            axis_value_ranges = dict()

            for value, value_name in best_axis.values.items():
                min_vp = 1e9
                max_vp = 0
                found_any = False
                for variant, price in details.variants.items():
                    if variant[best_axis.code_name] != value:
                        continue
                    max_vp = max(max_vp, price)
                    min_vp = min(min_vp, price)
                    found_any = True

                if found_any:
                    axis_value_ranges[value_name] = (int(min_vp), int(max_vp))

            axes_ranges[item_id] = (best_axis.code_name, axis_value_ranges)

    return await render_template(
        "search.html.j2",
        search_form=search_form,
        results=results,
        axes_ranges=axes_ranges,
        axes_errors=axes_errors,
    )
